from perivale import Buffer


def in_range(character: str, lower: str, upper: str) -> bool:
    character = ord(character)
    lower = ord(lower)
    upper = ord(upper)

    if lower >= upper:
        raise ValueError(f"illogical range ({lower} >= {upper}")

    return character >= lower and character <= upper


from .in_range import in_range
from .parse_escape_code import parse_escape_code
from .parse_integer import parse_integer, integer_present
from .parse_identifier import parse_identifier, identifier_present


__all__ = [
    "in_range",
    "identifier_present",
    "integer_present",
    "parse_escape_code",
    "parse_integer",
    "parse_identifier",
]


from perivale import Buffer


def _is_identifier(character: str, first_letter: bool) -> bool:
    ascii = ord(character)
    return (character == "_"
            or (ascii >= ord("a") and ascii <= ord("z"))
            or (ascii >= ord("A") and ascii <= ord("Z"))
            or (not first_letter and ascii >= ord("0") and ascii <= ord("9")))


def identifier_present(buffer: Buffer) -> bool:
    """ Checks whether an identifier is present
    
    An identifier is a string consisting of underscores, lower or uppercase 
    letters, and (non-leading) digits
    
    Arguments
    ---------
    buffer: Buffer
        the buffer to check in
    
    Returns
    -------
    present: bool
        if an identifier was found
    """

    return not buffer.finished() and _is_identifier(buffer.read(), True)


def parse_identifier(buffer: Buffer, consume: bool = False) -> str:
    """ Parses an identifier
    
    An identifier is a string consisting of underscores, lower or uppercase 
    letters, and (non-leading) digits
    
    Arguments
    ---------
    buffer: Buffer
        the buffer to parse from
    consume: bool
        if True, consumes the identifier
    
    Returns
    -------
    identifier: str
        the value of the identifier found
    
    Raises
    ------
    no_identifier: Exception
        if no identifier was found
    """

    text = ""
    position = buffer.copy_position()
    first_letter = True
    while not buffer.finished():

        character = buffer.read()
        if not _is_identifier(character, first_letter):
            break
        text += character
        buffer.increment()
        first_letter = False

    if not text:
        raise Exception("no identifier found")
    
    if not consume:
        buffer.position = position
    
    return text


from perivale import Buffer, ParseError


def parse_escape_code(buffer: Buffer, 
        escape_codes: dict, 
        errors: list = [],
        consume: bool = False) -> str | None:
    
    # Sort the codes so none get dropped
    sorted_codes = []
    for symbol, code in escape_codes.items():

        if not symbol:
            raise ValueError("empty symbol")
        elif symbol == "\\" or symbol[0] != "\\":

            # Add the backslash prefix, check that wasn't already in the dict
            symbol = f"\\{symbol}"
            if symbol in escape_codes:
                raise ValueError(f"duplicate escape code: '{symbol}'")

        # Insert with precedence
        for index in range(len(sorted_codes)):
            predecessor = sorted_codes[index][0]
            if symbol[:len(predecessor)] == predecessor:
                sorted_codes.insert(index, (symbol, code))
                break
        else:
            sorted_codes.append((symbol, code))
    
    # Check for a match with each pair
    for pair in sorted_codes:
        symbol, code = pair
        if buffer.match(symbol, consume=consume):
            return code
    
    # Check a code was, indeed, present
    if not buffer.match("\\"):
        raise Exception("expected an escape code")
    
    code_position = buffer.copy_position()
    buffer.increment()

    # Report unfinished/invalid code
    if buffer.finished() or buffer.match("\n"):
        error = ParseError()
        error.add_excerpt("incomplete escape code", 
                buffer.excerpt(code_position))
        errors.append(error)
        return None
    else:
        error = ParseError()
        error.add_excerpt("invalid escape code", 
                buffer.excerpt(code_position))
        errors.append(error)
        return None


from .buffer import Buffer
from .errors import ParseError
from .excerpt import Excerpt
from .position import Position

from .helpers import *

__all__ = [
    "Buffer", 
    "ParseError", 
    "Excerpt",
    "Position",
]


from perivale import Buffer


def _is_decimal(character: str) -> bool:
        character = ord(character)
        return character >= ord("0") and character <= ord("9")

  
def _is_hexadecimal(character: str) -> bool:
    character = ord(character)
    return ((character >= ord("0") and character <= ord("9"))
            or (character >= ord("a") and character <= ord("f"))
            or (character >= ord("A") and character <= ord("F")))


def _is_octal(character: str) -> bool:
    character = ord(character)
    return character >= ord("0") and character <= ord("7")


def _is_binary(character: str) -> bool:
    return character == "0" or character == "1"


def integer_present(buffer: Buffer, base: int = 10) -> bool:
    """ Checks whether an integer value of a given base is present

    Bases supported:

        10: decimal
        16: hexadecimal
        8: octal
        2: binary
    
    Arguments
    ---------
    buffer: Buffer
        buffer to check in
    base: int
        the base of the integer (defaults to 10, decimal)
    
    Returns
    -------
    present: bool
        True if an integer of that base is present
    
    Raises
    ------
    bad_base: ValueError
        if the base specified isn't supported
    """

    base_validators = {
        10: _is_decimal,
        16: _is_hexadecimal,
        2: _is_binary,
        8: _is_octal,
    }

    if base not in base_validators:
        raise ValueError(f"unsupported base: {base}")
    validator = base_validators[base]

    return not buffer.finished() and validator(buffer.read())


def parse_integer(buffer: Buffer, 
        base: int = 10, 
        consume: bool = False) -> int:
    
    """ Parses an integer value of a given base

    Bases supported:

        10: decimal
        16: hexadecimal
        8: octal
        2: binary
    
    Arguments
    ---------
    buffer: Buffer
        buffer to check in
    base: int
        the base of the integer (defaults to 10, decimal)
    consume: bool
        if True, consumes the integer
    
    Returns
    -------
    value: int
        the parsed value, cast to an integer
    
    Raises
    ------
    no_integer: Exception
        if no integer was found (check first)
    bad_base: ValueError
        if the base specified isn't supported
    """

    base_validators = {
        10: _is_decimal,
        16: _is_hexadecimal,
        2: _is_binary,
        8: _is_octal,
    }

    if base not in base_validators:
        raise ValueError(f"unsupported base: {base}")
    validator = base_validators[base]

    text = ""
    position = buffer.copy_position()
    while not buffer.finished():

        character = buffer.read()
        if not validator(character):
            break
        text += character
        buffer.increment()

    if not text:
        raise Exception("no integer found")
    
    if not consume:
        buffer.position = position
    
    return int(text, base)


from __future__ import annotations


class Position:

    def __init__(self, index: int = 0, line: int = 1, column: int = 1):
        self.index = index
        self.line = line
        self.column = column
    
    def __str__(self) -> str:
        return f"[{self.line}:{self.column}]"


from __future__ import annotations

from .position import Position


class Excerpt:

    def __init__(self, buffer, position: Position):
        if not buffer.position_valid(position):
            raise ValueError("invalid position")

        self.source = buffer.source
        self.text = buffer.line_text(position.line)
        self.position = position
    
    def __str__(self):
        result = f"{self.position}"

        if self.source:
            result += f" ({self.source})"
        
        result += f"\n{self.text}"

        line_length = len(self.text)
        column = self.position.column
        caret_index = column - 1 if column != -1 else line_length
        caret = " " * caret_index + "^"
        result += f"\n{caret}"

        return result



from __future__ import annotations

from .excerpt import Excerpt


class ParseError:

    def __init__(self):
        self.excerpts = []
    
    def __str__(self) -> str:

        result = []
        for pair in self.excerpts:
            message, excerpt = pair
            annotated_excerpt = message

            lines = f"{excerpt}".split("\n")
            annotated_excerpt += "".join([f"\n\t{line}" for line in lines])
            result.append(annotated_excerpt)
        
        return "\n".join(result)
    
    def add_excerpt(self, message: str, excerpt: Excerpt):
        self.excerpts.append((message, excerpt))


from copy import copy

from .position import Position
from .excerpt import Excerpt


class Buffer:

    def __init__(self, text: str, source: str = ""):
        self.text = text
        self.length = len(text)
        self.source = source

        self.line_indices = []
        self.line_indentations = []
        self.line_count = 0
        
        index = 0
        while index < self.length:
            self.line_indices.append(index)
        
            # Evaluate indentation
            indentation = 0
            while True:
                if index == self.length:
                    break
            
                if text[index] == "\t":
                    indentation = ((indentation + 4) & ~0x03)
                elif text[index] == " ":
                    indentation += 1
                else:
                    break
            
                index += 1
            self.line_indentations.append(indentation)

            # Move to next newline
            while index < self.length and text[index] != '\n':
                index += 1
            index += 1
        
        self.line_count = len(self.line_indices)
        
        self.position = Position()
        if not text:
            self.position.line = -1
            self.position.column = -1
        elif text[0] == "\n":
            self.position.column = -1
    
    def position_valid(self, position: Position):
        """ Checks whether a given position is valid for this buffer
        
        Arguments
        ---------
        position: Position
            the position to check
        
        Returns
        -------
        valid: bool
            true if the position is valid
        """

        index, line, column = position.index, position.line, position.column

        if index == self.length and line == -1 and column == -1:
            return True

        if index > self.length or line > self.line_count:
            return False
        
        start = self.line_indices[line - 1]
        if line == self.line_count and column == -1:
            return index == self.length - 1
        
        if column == -1:
            if line == self.line_count:
                return index == self.length
            
            next_index = self.line_indices[line]
            return index == next_index - 1
        
        return index == start + column - 1        

    def increment(self, steps: int = 1):
        """ Increments the buffer's position
        
        Arguments
        ---------
        steps: int
            the number of steps to increment
        """

        for _ in range(steps):

            # Stop when the end of the buffer has been reached
            if self.position.index >= self.length:
                return

            # If the current character is a newline, increment the line number
            # and reset the column
            character = self.text[self.position.index]
            if character == "\n":
                self.position.line += 1
                self.position.column = 1

            # Handle tabs by rounding up to the next multiple of 4 (+1)
            elif character == "\t":
                column = int(((self.position.column + 3) / 4) * 4 + 1)
                self.position.column = column

            # Failing the above, and if the character wasn't format-related,
            # increment the column
            else:

                value = ord(character)
                lower, upper = [ord(letter) for letter in [" ", "~"]]
                if value >= lower and value <= upper:
                    self.position.column += 1

            self.position.index += 1

            if self.position.index == self.length:
                self.position.column = -1
                self.position.line = -1
            elif self.text[self.position.index] == "\n":
                self.position.column = -1
    
    def finished(self) -> bool:
        """ Checks if the buffer is finished
        
        Returns
        -------
        finished: bool
            if the buffer is finished
        """
        
        return self.position.index == self.length
    
    def copy_position(self) -> Position:
        """ Copies the buffer's position
        
        Note: this method _must_ be used rather than using the position member 
        directly
        
        Returns
        -------
        position: Position
            the copied position
        """

        return copy(self.position)
    
    def set_position(self, position: Position):
        """ Sets the buffer's position, checking its validity
        
        Arguments
        ---------
        position: Position
            the position to set
        
        Raises
        ------
        error: IndexError
            if the position is invalid
        """

        if not self.position_valid(position):
            raise IndexError(f"invalid position: {position}")
        self.position = copy(position)
    
    def read(self, consume: bool = False) -> str:
        """ Reads the next character from the buffer
        
        Arguments
        ---------
        consume: bool
            if True, increments past the read character
            
        Returns
        -------
        character: str
            the character read, or "" if the buffer is finished
        """

        if self.finished():
            return ""
        
        result = self.text[self.position.index]
        if consume:
            self.increment()
        return result
    
    def match(self, text: str, consume: bool = False) -> bool:
        """ Checks whether a substring matches
        
        Arguments
        ---------
        text: str
            the substring to match
        consume: bool
            if true (and the substring matches), increments past it
        
        Returns
        -------
        match: bool
            true if the substring matched
        """

        if self.finished():
            return False
        
        # Check match feasible
        length = len(text)
        if self.position.index + length > self.length:
            return False
        
        # Extract substring
        end = self.position.index + length
        substring = self.text[self.position.index:end]

        # Check for match
        result = text == substring
        if result and consume:
            self.increment(steps=length)
        
        return result

    def skip_line(self):
        """ Skips the current line """

        if self.position.line >= self.line_count:
            self.position.index = self.length
            self.position.column = -1
        else:
            index = self.line_indices[self.position.line]
            self.position.index = index
            self.position.line += 1
            self.position.column = 1

    def skip_space(self, include_newlines: bool = False):
        """ Skips whitespace characters
        
        Arguments
        ---------
        include_newlines: bool
            if true, skips newlines as well
        """
        while not self.finished():
            character = self.text[self.position.index]

            if ((character not in " \t\v\r") and 
                    not (character == "\n" and include_newlines)):
                break

            self.increment()

    def line_text(self, line_number: int = 0) -> str:
        """ Gets the text of a given line
        
        Arguments
        ---------
        line_number: int
            the number of the line of text to fetch; Note: not zero-indexed! By
            default, gets the current line's text
        
        Returns
        -------
        text: str
            the given line's text
        """

        # Set line number to current if zero, or last line if -1
        if line_number == 0:
            line_number = self.position.line
        elif line_number == -1:
            line_number = self.line_count

        # Assert line number in range
        if line_number < 1 or line_number > self.line_count:
            return ""

        # Evaluate start index
        start_index = 0
        if line_number > 1:
            start_index = self.line_indices[line_number - 1]

        # Evaluate end index
        end_index = self.length
        if line_number < self.line_count:
            end_index = self.line_indices[line_number] - 1
        elif self.length and self.text[self.length - 1] == "\n":
            end_index = self.length - 1

        return self.text[start_index:end_index]

    def line_indentation(self, line_number: int = 0) -> int:
        """ Gets the indentation of a given line
        
        Arguments
        ---------
        line_number: int
            the number of the line whose indentation to fetch; Note: not 
            zero-indexed! By default, gets the current line's indentation
        
        Returns
        -------
        indentation: int
            the given line's indentation
        """

        # Set line number to current if zero, or last line if -1
        if line_number == 0:
            line_number = self.position.line
        elif line_number == -1:
            line_number = self.line_count

        # Check line is in bounds
        if line_number < 1 or line_number > self.line_count:
            return 0

        return self.line_indentations[line_number - 1]
    
    def excerpt(self, position: Position = None) -> Excerpt:
        """ Creates an excerpt

        If no position is given, the current buffer position is inferred
        
        Arguments
        ---------
        position: Position
            the position of the excerpt
        
        Returns
        -------
        excerpt: Excerpt
            the excerpt created
        
        Raises
        ------
        position_invalid: ValueError
            if the position given is invalid
        """

        if not position:
            position = self.copy_position()
        return Excerpt(self, position)